from __future__ import annotations

from typing import Any, Callable, Dict, List, Optional, Set, Union

from supervisely.app import DataJson, StateJson
from supervisely.app.widgets import Widget


class Collapse(Widget):
    """A Collapse widget allows users to efficiently manage and navigate content by toggling between
    hidden and visible states, promoting a more compact and organized user interface.

    Read about it in `Developer Portal <https://developer.supervisely.com/app-development/widgets/layouts-and-containers/collapse>`_
        (including screenshots and examples).

    :param items: list of items to be displayed in the widget
    :type items: List[Collapse.Item]
    :param accordion: if True, only one panel could be active at a time
    :type accordion: bool
    :param widget_id: An identifier of the widget.
    :type widget_id: str, optional

    :Usage example:
        .. code-block:: python

        from supervisely.app.widgets import Collapse

        items = [
            Collapse.Item("item1", "Item 1", "Content of item 1"),
            Collapse.Item("item2", "Item 2", "Content of item 2"),
        ]
        collapse = Collapse(items)

    """

    class Routes:
        VALUE_CHANGED = "value_changed"

    class Item(object):
        """Represents an item of Collapse widget.

        :param name: unique identification of the panel
        :type name: str
        :param title: title of the panel
        :type title: str
        :param content: content of the panel
        :type content: Optional[Union[Widget, str]]
        """

        def __init__(self, name: str, title: str, content: Optional[Union[Widget, str]]):
            self.name = name
            self.title = title
            self.content = content

        def to_json(self) -> Dict[str, Any]:
            """Returns JSON representation of the item.
            Dictionary contains the following keys:
                - name: unique identification of the panel
                - label: title of the panel
                - content_type: type of the content (str or Widget)

            :return: JSON representation of the item
            :rtype: Dict[str, Any]
            """
            if isinstance(self.content, str):
                content_type = "text"
            else:
                content_type = str(type(self.content))
            return {
                "name": self.name,
                "label": self.title,
                "content_type": content_type,
            }

    def __init__(
        self,
        items: Optional[List[Collapse.Item]] = None,
        accordion: Optional[bool] = False,
        widget_id: Optional[str] = None,
    ):
        if items is None:
            items = [Collapse.Item("default", "Empty default item", "")]

        labels = [item.name for item in items]
        if len(set(labels)) != len(labels):
            raise ValueError("All items must have a unique name.")

        self._items: List[Collapse.Item] = items

        self._accordion = accordion
        self._active_panels = []

        self._items_names = set(labels)
        super().__init__(widget_id=widget_id, file_path=__file__)

    def _get_items_json(self) -> List[Dict[str, Any]]:
        return [item.to_json() for item in self._items]

    def get_json_data(self) -> Dict[str, Any]:
        """Returns dictionary with widget data, which defines the appearance and behavior of the widget.

        Dictionary contains the following fields:
            - accordion: if True, only one panel could be active at a time
            - items: list of items to be displayed in the widget

        :return: dictionary with widget data
        :rtype: Dict[str, Any]
        """
        return {
            "accordion": self._accordion,
            "items": self._get_items_json(),
        }

    def get_json_state(self) -> Dict[str, List[str]]:
        """Returns dictionary with widget state.

        Dictionary contains the following fields:
            - value: list of active panels

        :return: dictionary with widget state
        :rtype: Dict[str, List[str]]
        """
        return {"value": self._active_panels}

    def set_active_panel(self, value: Union[str, List[str]]) -> None:
        """Set active panel or panels.

        :param value: panel name(s)
        :type value: Union[str, List[str]]
        :raises TypeError: value of type List[str] can't be setted, if accordion is True.
        :raises ValueError: panel with such title doesn't exist.
        """
        if isinstance(value, list):
            if self._accordion:
                raise TypeError(
                    "Only one panel could be active in accordion mode. Use `str`, not `list`."
                )
            for name in value:
                if name not in self._items_names:
                    raise ValueError(
                        f"Can't activate panel `{name}`: item with such name doesn't exist."
                    )
        else:
            if value not in self._items_names:
                raise ValueError(
                    f"Can't activate panel `{value}`: item with such name doesn't exist."
                )

        if isinstance(value, str):
            self._active_panels = [value]
        else:
            self._active_panels = value

        StateJson()[self.widget_id]["value"] = self._active_panels
        StateJson().send_changes()

    def get_active_panel(self) -> Union[str, List[str]]:
        """Returns active panel or panels.

        :return: active panel or panels
        :rtype: Union[str, List[str]]
        """
        return StateJson()[self.widget_id]["value"]

    def get_items(self) -> List[Collapse.Item]:
        """Returns list of items.

        :return: list of items
        :rtype: List[Collapse.Item]
        """
        return DataJson()[self.widget_id]["items"]

    def set_items(self, value: List[Collapse.Item]) -> None:
        """Set items for the widget.
        This method replaces all existing items with new ones.
        To add new items, use :meth:`add_items` method.

        :param value: list of items to be displayed in the widget
        :type value: List[Collapse.Item]
        """
        names = [val.name for val in value]

        self._items_names = self._make_set_from_unique(names)
        self._items = value
        self._active_panels = []

        DataJson()[self.widget_id]["items"] = self._get_items_json()
        DataJson().send_changes()

    def add_items(self, value: List[Collapse.Item]) -> None:
        """Add items to the widget.
        This method adds new items to the existing ones.
        To replace all existing items with new ones, use :meth:`set_items` method.

        :param value: list of items to be displayed in the widget
        :type value: List[Collapse.Item]
        :raises ValueError: item with such name already exists.
        """
        names = [val.name for val in value]
        set_of_names = self._make_set_from_unique(names)

        for name in names:
            if name in self._items_names:
                raise ValueError(f"Item with name {name} already exists.")

        self._items.extend(value)
        self._items_names.update(set_of_names)
        DataJson()[self.widget_id]["items"] = self._get_items_json()
        DataJson().send_changes()

    def value_changed(self, func: Callable[[List[str]], Any]) -> Callable[[], None]:
        """Decorator for the function to be called when active panel or panels are changed.

        :param func: function to be called when active panel or panels are changed
        :type func: Callable[[List[str]], Any]
        :return: decorated function
        :rtype: Callable[[], None]
        """
        route_path = self.get_route_path(Collapse.Routes.VALUE_CHANGED)
        server = self._sly_app.get_server()
        self._changes_handled = True

        @server.post(route_path)
        def _click():
            active = self.get_active_panel()
            self._active_panels = active
            func(active)

        return _click

    @property
    def items_names(self) -> Set[str]:
        """Returns set of names of items.

        :return: set of names of items
        :rtype: Set[str]
        """

        return self._items_names

    def _make_set_from_unique(self, names: List[str]) -> Set[str]:
        set_of_names = set(names)
        if len(names) != len(set_of_names):
            raise ValueError("All items must have a unique name.")
        return set_of_names



from concurrent.futures import ThreadPoolExecutor
from dataclasses import MISSING, dataclass, fields
from json import JSONDecodeError
from os.path import dirname, join
from typing import Dict, List, Optional, Union

import requests

from supervisely import logger
from supervisely.api.api import Api, ApiField

EXPERIMENT_INFO_FILENAME = "experiment_info.json"


@dataclass
class ExperimentInfo:
    experiment_name: str
    """Name of the experiment. Defined by the user in the training app"""
    framework_name: str
    """Name of the framework used in the experiment"""
    model_name: str
    """Name of the model used in the experiment. Defined by the user in the training app"""
    task_type: str
    """Task type of the experiment"""
    project_id: int
    """Project ID in Supervisely"""
    task_id: int
    """Task ID in Supervisely"""
    model_files: dict
    """Dictionary with paths to model files that needs to be downloaded for training"""
    model_meta: str
    """Path to file with model metadata such as model name, project id, project name and classes used for training"""
    checkpoints: List[str]
    """List of relative paths to checkpoints"""
    best_checkpoint: str
    """Name of the best checkpoint. Defined by the user in the training app"""
    hyperparameters: str
    """Path to .yaml file with hyperparameters used in the experiment"""
    artifacts_dir: str
    """Path to the directory with artifacts"""
    base_checkpoint: Optional[str] = None
    """Name of the base checkpoint used for training"""
    base_checkpoint_link: Optional[str] = None
    """Link to the base checkpoint used for training. URL in case of pretrained model, or Team Files path in case of custom model."""
    export: Optional[dict] = None
    """Dictionary with exported weights in different formats"""
    app_state: Optional[str] = None
    """Path to file with settings that were used in the app"""
    train_val_split: Optional[str] = None
    """Path to train and validation splits, which contains IDs of the images used in each split"""
    train_size: Optional[int] = None
    """Number of images in the training set"""
    val_size: Optional[int] = None
    """Number of images in the validation set"""
    datetime: Optional[str] = None
    """Date and time when the experiment was started"""
    evaluation_report_id: Optional[int] = None
    """ID of the model benchmark evaluation report"""
    evaluation_report_link: Optional[str] = None
    """Link to the model benchmark evaluation report"""
    evaluation_metrics: Optional[dict] = None
    """Evaluation metrics"""
    logs: Optional[dict] = None
    """Dictionary with link and type of logger"""

    def __init__(self, **kwargs):
        required_fieds = {
            field.name for field in fields(self.__class__) if field.default is MISSING
        }
        missing_fields = required_fieds - set(kwargs.keys())
        if missing_fields:
            raise ValueError(
                f"ExperimentInfo missing required arguments: '{', '.join(missing_fields)}'"
            )
        field_names = set(f.name for f in fields(self.__class__))
        kwargs = {k: v for k, v in kwargs.items() if k in field_names}
        for key, value in kwargs.items():
            setattr(self, key, value)

    def to_json(self) -> Dict:
        data = {}
        for field in fields(self.__class__):
            value = getattr(self, field.name)
            data[field.name] = value


def get_experiment_infos(api: Api, team_id: int, framework_name: str) -> List[ExperimentInfo]:
    """
    Get experiments from the specified framework folder for Train v2

    :param api: Supervisely API client
    :type api: Api
    :param team_id: Team ID
    :type team_id: int
    :param framework_name: Name of the framework
    :type framework_name: str
    :return: List of ExperimentInfo objects
    :rtype: List[ExperimentInfo]
    :Usage example:

     .. code-block:: python

        import supervisely as sly

        api = sly.Api.from_env()
        team_id = sly.env.team_id()
        framework_name = "RT-DETRv2"
        experiment_infos = sly.nn.training.experiments.get_experiment_infos(api, team_id, framework_name)
    """
    metadata_name = "experiment_info.json"
    experiments_folder = "/experiments"
    experiment_infos = []

    file_infos = api.file.list(team_id, experiments_folder, recursive=True, return_type="fileinfo")
    sorted_experiment_paths = []
    for file_info in file_infos:
        if not file_info.path.endswith(metadata_name):
            continue

        experiment_dir = dirname(file_info.path)
        if experiment_dir.endswith(framework_name):
            experiment_path = join(experiment_dir, metadata_name)
            sorted_experiment_paths.append(experiment_path)

    def fetch_experiment_data(experiment_path: str):
        try:
            response = api.post(
                "file-storage.download",
                {ApiField.TEAM_ID: team_id, ApiField.PATH: experiment_path},
                stream=True,
            )
            response.raise_for_status()
            response_json = response.json()
            required_fields = {
                field.name for field in fields(ExperimentInfo) if field.default is not None
            }
            optional_fields = {
                field.name for field in fields(ExperimentInfo) if field.default is None
            }

            missing_optional_fields = optional_fields - response_json.keys()
            if missing_optional_fields:
                logger.debug(
                    f"Missing optional fields: {missing_optional_fields} for '{experiment_path}'"
                )
                for field in missing_optional_fields:
                    response_json[field] = None

            missing_required_fields = required_fields - response_json.keys()
            if missing_required_fields:
                logger.debug(
                    f"Missing required fields: {missing_required_fields} for '{experiment_path}'. Skipping."
                )
                return None
            field_names = {field.name for field in fields(ExperimentInfo)}
            return ExperimentInfo(**{k: v for k, v in response_json.items() if k in field_names})
        except requests.exceptions.RequestException as e:
            logger.debug(f"Request failed for '{experiment_path}': {e}")
        except JSONDecodeError as e:
            logger.debug(f"JSON decode failed for '{experiment_path}': {e}")
        except TypeError as e:
            logger.error(f"TypeError for '{experiment_path}': {e}")
        return None

    with ThreadPoolExecutor() as executor:
        experiment_infos = list(executor.map(fetch_experiment_data, sorted_experiment_paths))

    experiment_infos = [info for info in experiment_infos if info is not None]
    return experiment_infos


def _fetch_experiment_data(api, team_id: int, experiment_path: str) -> Union[ExperimentInfo, None]:
    """
    Fetch experiment data from the specified path in Supervisely Team Files

    :param api: Supervisely API client
    :type api: Api
    :param team_id: Team ID
    :type team_id: int
    :param experiment_path: Path to the experiment data
    :type experiment_path: str
    :return: ExperimentInfo object
    :rtype: Union[ExperimentInfo, None]
    """
    try:
        response = api.post(
            "file-storage.download",
            {ApiField.TEAM_ID: team_id, ApiField.PATH: experiment_path},
            stream=True,
        )
        response.raise_for_status()
        response_json = response.json()
        required_fields = {
            field.name for field in fields(ExperimentInfo) if field.default is MISSING
        }
        optional_fields = {
            field.name for field in fields(ExperimentInfo) if field.default is not MISSING
        }

        missing_optional_fields = optional_fields - response_json.keys()
        if missing_optional_fields:
            logger.debug(
                f"Missing optional fields: {missing_optional_fields} for '{experiment_path}'"
            )
            for field in missing_optional_fields:
                response_json[field] = None

        missing_required_fields = required_fields - response_json.keys()
        if missing_required_fields:
            logger.debug(
                f"Missing required fields: {missing_required_fields} for '{experiment_path}'. Skipping."
            )
            return None
        all_fields = required_fields | optional_fields
        return ExperimentInfo(**{k: v for k, v in response_json.items() if k in all_fields})
    except requests.exceptions.RequestException as e:
        logger.debug(f"Request failed for '{experiment_path}': {e}")
    except JSONDecodeError as e:
        logger.debug(f"JSON decode failed for '{experiment_path}': {e}")
    except TypeError as e:
        logger.error(f"TypeError for '{experiment_path}': {e}")
    return None


def get_experiment_info_by_artifacts_dir(
    api: Api, team_id: int, artifacts_dir: str
) -> Union[ExperimentInfo, None]:
    """
    Get experiment info by artifacts directory

    :param api: Supervisely API client
    :type api: Api
    :param team_id: Team ID
    :type team_id: int
    :param artifacts_dir: Path to the directory with artifacts
    :type artifacts_dir: str
    :return: ExperimentInfo object
    :rtype: Optional[ExperimentInfo]
    :Usage example:

     .. code-block:: python

        import supervisely as sly

        api = sly.Api.from_env()
        team_id = sly.env.team_id()
        artifacts_dir = "/experiments/27_Lemons (Rectangle)/265_RT-DETRv2/"
        experiment_info = sly.nn.training.experiments.get_experiment_info_by_artifacts_dir(api, team_id, artifacts_dir)
    """
    if not artifacts_dir.startswith("/experiments"):
        raise ValueError("Artifacts directory should start with '/experiments'")
    experiment_path = join(artifacts_dir, EXPERIMENT_INFO_FILENAME)
    return _fetch_experiment_data(api, team_id, experiment_path)



# coding: utf-8

from supervisely.api.entity_annotation.object_api import ObjectApi
from supervisely.video_annotation.key_id_map import KeyIdMap
from supervisely.volume_annotation.volume_object_collection import (
    VolumeObjectCollection,
)


class VolumeObjectApi(ObjectApi):
    """
    :class:`VolumeObject<supervisely.volume_annotation.volume_object.VolumeObject>` for :class:`VolumeAnnotation<supervisely.volume_annotation.volume_annotation.VolumeAnnotation>`.
    """

    def append_bulk(
        self, volume_id: int, objects: VolumeObjectCollection, key_id_map: KeyIdMap = None
    ):
        """
        Add Tags to Annotation Objects

        :param volume_id: Volume ID in Supervidely.
        :type volume_id: int
        :param objects: VolumeAnnotation objects.
        :type objects: VolumeObjectCollection
        :param key_id_map: KeyIdMap object.
        :type key_id_map: KeyIdMap, optional
        :return: List of objects IDs
        :rtype: :class:`List[int]`

        :Usage example:

         .. code-block:: python

            import supervisely as sly

            from supervisely.volume_annotation.volume_tag import VolumeTag
            from supervisely.video_annotation.key_id_map import KeyIdMap

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            project_id = 17209
            volume_id = 19402023

            meta_json = api.project.get_meta(project_id)
            project_meta = sly.ProjectMeta.from_json(meta_json)

            key_id_map = KeyIdMap()
            ann_info = api.volume.annotation.download(volume_id)
            ann = sly.VolumeAnnotation.from_json(ann_info, project_meta, key_id_map)

            api.volume.object.append_bulk(volume_id, ann.objects, key_id_map)
        """

        info = self._api.volume.get_info_by_id(volume_id)
        return self._append_bulk(
            self._api.volume.tag,
            volume_id,
            info.project_id,
            info.dataset_id,
            objects,
            key_id_map,
        )



import asyncio
import os
import time

from tqdm import tqdm
from tqdm.asyncio import tqdm

import supervisely as sly

LOG_LEVEL = "INFO"
# LOG_LEVEL = "DEBUG"
PROJECT_ID = 42569
user_path = os.path.expanduser("~")
save_path = f"{user_path}/Work/test_volumes_download/"
sly.logger.info(f"Save path: {save_path}")
sly.fs.ensure_base_path(save_path)
sly.fs.clean_dir(save_path)
api = sly.Api.from_env()
datasets = api.dataset.get_list(PROJECT_ID, recursive=True)
volumes = []
for dataset in datasets:
    items = api.volume.get_list(dataset.id)
    volumes.extend(items)
ids = [volume.id for volume in volumes]
names = [volume.name for volume in volumes]
paths = [f"{save_path}{volume.name}" for volume in volumes]

api.logger.setLevel(LOG_LEVEL)


def main_dps():
    pbar = sly.tqdm_sly(total=len(ids), desc="Downloading volumes", unit="volume")
    loop = asyncio.get_event_loop()
    loop.run_until_complete(api.volume.download_paths_async(ids, paths, progress_cb=pbar))


def compare_main_dps():
    start = time.time()
    for id, path in zip(ids, paths):
        pbar = tqdm(desc="Downloading volume", unit="B", unit_scale=True)
        api.volume.download_path(id, path, pbar)
        pbar.close()

    finish = time.time() - start
    print(f"Time taken for old method: {finish}")

    start = time.time()
    main_dps()
    finish = time.time() - start
    print(f"Time taken for async method: {finish}")


if __name__ == "__main__":
    try:
        # main_dps()  # to download and save volumes as files (batch)
        compare_main_dps()  # to compare the time taken for downloading volumes as files (batch)

    except KeyboardInterrupt:
        sly.logger.info("Stopped by user")



from os.path import join
from typing import Any, Dict, List, Optional, Union

import yaml

import supervisely.io.env as sly_env
import supervisely.io.fs as sly_fs
import supervisely.io.json as sly_json
from supervisely import Api
from supervisely.app.widgets import Card, Container, Field, RadioTabs, SelectString, Text, Widget
from supervisely.app.widgets.experiment_selector.experiment_selector import ExperimentSelector
from supervisely.app.widgets.pretrained_models_selector.pretrained_models_selector import (
    PretrainedModelsSelector,
)
from supervisely.nn.experiments import get_experiment_infos
from supervisely.nn.inference.gui.serving_gui import ServingGUI
from supervisely.nn.utils import ModelSource, RuntimeType, _get_model_name


class ServingGUITemplate(ServingGUI):
    def __init__(
        self,
        framework_name: str,
        models: Optional[list] = None,
        app_options: Optional[str] = None,
    ):
        if not isinstance(framework_name, str):
            raise ValueError("'framework_name' must be a string")
        super().__init__()

        self.api = Api.from_env()
        self.team_id = sly_env.team_id()

        self.framework_name = framework_name
        self.models = models
        self.app_options = self._load_app_options(app_options) if app_options else {}

        base_widgets = self._initialize_layout()
        extra_widgets = self._initialize_extra_widgets()

        self.widgets = base_widgets + extra_widgets
        self.card = self._get_card()

    def _get_card(self) -> Card:
        return Card(
            title="Select Model",
            description="Select the model to deploy and press the 'Serve' button.",
            content=Container(widgets=self.widgets, gap=10),
            overflow="unset",
        )

    def _initialize_layout(self) -> List[Widget]:
        # Pretrained models
        use_pretrained_models = self.app_options.get("pretrained_models", True)
        use_custom_models = self.app_options.get("custom_models", True)

        if not use_pretrained_models and not use_custom_models:
            raise ValueError(
                "At least one of 'pretrained_models' or 'custom_models' must be enabled."
            )

        if use_pretrained_models and self.models is not None:
            self.pretrained_models_table = PretrainedModelsSelector(self.models)
        else:
            self.pretrained_models_table = None

        # Custom models
        if use_custom_models:
            experiments = get_experiment_infos(self.api, self.team_id, self.framework_name)
            self.experiment_selector = ExperimentSelector(self.team_id, experiments)
        else:
            self.experiment_selector = None

        # Tabs
        tabs = []
        if self.pretrained_models_table is not None:
            tabs.append(
                (
                    ModelSource.PRETRAINED,
                    "Publicly available models",
                    self.pretrained_models_table,
                )
            )
        if self.experiment_selector is not None:
            tabs.append(
                (
                    ModelSource.CUSTOM,
                    "Models trained in Supervisely",
                    self.experiment_selector,
                )
            )
        if tabs:
            titles, descriptions, content = zip(*tabs)
            self.model_source_tabs = RadioTabs(
                titles=titles,
                descriptions=descriptions,
                contents=content,
            )
        else:
            self.model_source_tabs = None

        # Runtime
        default_runtime = RuntimeType.PYTORCH
        available_runtimes = {
            value.lower(): value
            for name, value in vars(RuntimeType).items()
            if not name.startswith("__")  # exclude private attributes
        }
        supported_runtimes_input = self.app_options.get("supported_runtimes", [default_runtime])
        supported_runtimes = [
            available_runtimes[runtime.lower()]
            for runtime in supported_runtimes_input
            if runtime.lower() in available_runtimes
        ]

        if len(supported_runtimes) > 1:
            self.runtime_select = SelectString(supported_runtimes)
            runtime_field = Field(self.runtime_select, "Runtime", "Select a runtime for inference.")
        else:
            self.runtime_select = None
            runtime_field = None

        # Layout
        card_widgets = [self.model_source_tabs]
        if runtime_field is not None:
            card_widgets.append(runtime_field)

        # Runtime exported checkpoint message
        self._export_msg = Text("")
        self._export_msg.hide()
        card_widgets.append(self._export_msg)

        if self.runtime_select is not None:
            self.runtime_select.value_changed(lambda _: self._update_export_message())
        if self.experiment_selector is not None:
            self.experiment_selector.value_changed(lambda _: self._update_export_message())
            for task_type in self.experiment_selector.rows:
                for row in self.experiment_selector.rows[task_type]:
                    row.checkpoints_selector.value_changed(lambda _: self._update_export_message())
        if self.pretrained_models_table is not None:
            self.pretrained_models_table.model_changed(lambda _: self._update_export_message())

        if self.model_source_tabs is not None:
            self.model_source_tabs.value_changed(lambda _: self._update_export_message())

        self._update_export_message()

        return card_widgets

    def _initialize_extra_widgets(self) -> List[Widget]:
        return []

    @property
    def model_source(self) -> str:
        return self.model_source_tabs.get_active_tab()

    @property
    def model_info(self) -> Dict[str, Any]:
        return self._get_selected_row()

    @property
    def model_name(self) -> Optional[str]:
        if self.model_source == ModelSource.PRETRAINED:
            return _get_model_name(self.model_info)
        else:
            return self.model_info.get("model_name")

    @property
    def model_files(self) -> List[str]:
        if self.model_source == ModelSource.PRETRAINED:
            model_meta = self.model_info.get("meta", {})
            return model_meta.get("model_files", {})
        else:
            return self.experiment_selector.get_model_files()

    @property
    def runtime(self) -> str:
        if self.runtime_select is not None:
            return self.runtime_select.get_value()
        return RuntimeType.PYTORCH

    def get_params_from_gui(self) -> Dict[str, Any]:
        return {
            "model_source": self.model_source,
            "model_files": self.model_files,
            "model_info": self.model_info,
            "device": self.device,
            "runtime": self.runtime,
        }

    def _load_app_options(self, app_options: str = None) -> Dict[str, Any]:
        """
        Loads the app_options parameter to ensure it is in the correct format.
        """
        if app_options is None:
            return {}

        if isinstance(app_options, str):
            if sly_fs.file_exists(app_options) and sly_fs.get_file_ext(app_options) in [
                ".yaml",
                ".yml",
            ]:
                with open(app_options, "r") as file:
                    app_options = yaml.safe_load(file)
        else:
            raise ValueError(
                "Invalid app_options file provided. Please provide a valid '.yaml' or '.yml' file with app_options."
            )
        if not isinstance(app_options, dict):
            raise ValueError("app_options must be a dict")
        return app_options

    def _get_selected_row(self) -> Dict[str, Any]:
        if self.model_source == ModelSource.PRETRAINED and self.pretrained_models_table:
            return self.pretrained_models_table.get_selected_row()
        elif self.model_source == ModelSource.CUSTOM and self.experiment_selector:
            return self.experiment_selector.get_selected_experiment_info()
        return {}

    def _update_export_message(self):
        self._export_msg.hide()

        runtime = self.runtime
        non_conversion_runtimes = [RuntimeType.ONNXRUNTIME, RuntimeType.TENSORRT]
        if runtime not in non_conversion_runtimes:
            return

        if self.model_source == ModelSource.PRETRAINED:
            self._export_msg.set(
                "Checkpoint will be converted before deployment.",
                "info",
            )
            self._export_msg.show()
            return

        checkpoint_name = None
        if self.model_source == ModelSource.CUSTOM and self.experiment_selector is not None:
            selected_row = self.experiment_selector.get_selected_row()
            if selected_row is None:
                return
            checkpoint_name = selected_row.get_selected_checkpoint_name()
            if checkpoint_name is None:
                return

        model_info = self.model_info or {}
        export_info = model_info.get("export", {})
        available = False
        if isinstance(export_info, dict):
            for key in export_info.keys():
                if key.lower().startswith(runtime.lower()):
                    available = True
                    break
            if checkpoint_name != selected_row.best_checkpoint:
                available = False

        if available:
            self._export_msg.set(
                "Runtime checkpoint exists – no conversion needed.",
                "info",
            )
        else:
            self._export_msg.set(
                "Checkpoint will be converted before deployment.",
                "info",
            )
        self._export_msg.show()



import os
from pathlib import Path
from typing import Optional, List
from supervisely import PointcloudEpisodeAnnotation, ProjectMeta, is_development, logger, ObjClass, ObjClassCollection
from supervisely.geometry.cuboid_3d import Cuboid3d
from supervisely.api.api import Api, ApiField
from supervisely.convert.base_converter import AvailablePointcloudEpisodesConverters
from supervisely.convert.pointcloud_episodes.kitti_360.kitti_360_helper import *
from supervisely.convert.pointcloud_episodes.pointcloud_episodes_converter import PointcloudEpisodeConverter
from supervisely.io.fs import (
    file_exists,
    get_file_name,
    get_file_name_with_ext,
    list_files_recursively,
    silent_remove,
)
from supervisely.pointcloud_annotation.pointcloud_episode_frame_collection import PointcloudEpisodeFrameCollection
from supervisely.pointcloud_annotation.pointcloud_episode_object_collection import PointcloudEpisodeObjectCollection
from supervisely.pointcloud_annotation.pointcloud_episode_object import PointcloudEpisodeObject
from supervisely.pointcloud_annotation.pointcloud_episode_frame import PointcloudEpisodeFrame
from supervisely.pointcloud_annotation.pointcloud_figure import PointcloudFigure

class KITTI360Converter(PointcloudEpisodeConverter):

    class Item:

        def __init__(
            self,
            scene_name: str,
            frame_paths: List[str],
            ann_data: Annotation3D,
            poses_path: str,
            related_images: Optional[tuple] = None,
            custom_data: Optional[dict] = None,
        ):
            self._scene_name = scene_name
            self._frame_paths = frame_paths
            self._ann_data = ann_data
            self._poses_path = poses_path
            self._related_images = related_images or []

            self._type = "point_cloud_episode"
            self._custom_data = custom_data if custom_data is not None else {}

    def __init__(self, *args, **kwargs):
        self._calib_path = None
        super().__init__(*args, **kwargs)

    def __str__(self) -> str:
        return AvailablePointcloudEpisodesConverters.KITTI360

    @property
    def key_file_ext(self) -> str:
        return ".bin"

    def validate_format(self) -> bool:
        try:
            import kitti360scripts
        except ImportError:
            logger.warn("Please run 'pip install kitti360Scripts' to import KITTI-360 data.")
            return False

        self._items = []
        subdirs = os.listdir(self._input_data)
        if len(subdirs) == 1:
            self._input_data = os.path.join(self._input_data, subdirs[0])

        # * Get calibration path
        calib_dir = next(iter([(Path(path).parent).as_posix() for path in list_files_recursively(self._input_data, [".txt"], None, True) if Path(path).stem.startswith("calib")]), None)
        if calib_dir is None:
            return False
        self._calib_path = calib_dir

        # * Get pointcloud files paths
        velodyne_files = list_files_recursively(self._input_data, [".bin"], None, True)
        if len(velodyne_files) == 0:
            return False

        # * Get annotation files paths and related images
        boxes_ann_files = list_files_recursively(self._input_data, [".xml"], None, True)
        if len(boxes_ann_files) == 0:
            return False
        rimage_files = list_files_recursively(self._input_data, [".png"], None, True)

        kitti_anns = []
        for ann_file in boxes_ann_files:
            key_name = Path(ann_file).stem

            # * Get pointcloud files
            frame_paths = []
            for path in velodyne_files:
                if key_name in Path(path).parts:
                    frame_paths.append(path)
            if len(frame_paths) == 0:
                logger.warn("No frames found for name: %s", key_name)
                continue

            # * Get related images
            rimages = []
            for rimage in rimage_files:
                path = Path(rimage)
                if key_name in path.parts:
                    cam_name = path.parts[-3]
                    rimages.append((cam_name, rimage))

            # * Get poses
            poses_filter = (
                lambda x: x.endswith("cam0_to_world.txt") and key_name in Path(x).parts
            )
            poses_path = next(
                path
                for path in list_files_recursively(self._input_data, [".txt"], None, True)
                if poses_filter(path)
            )
            if poses_path is None:
                logger.warn("No poses found for name: %s", key_name)
                continue

            # * Parse annotation
            ann = Annotation3D(ann_file)
            kitti_anns.append(ann)

            self._items.append(
                self.Item(key_name, frame_paths, ann, poses_path, rimages)
            )

        # * Get object class names for meta
        obj_class_names = set()
        for ann in kitti_anns:
            for obj in ann.get_objects():
                obj_class_names.add(obj.name)
        obj_classes = [ObjClass(obj_class, Cuboid3d) for obj_class in obj_class_names]
        self._meta = ProjectMeta(obj_classes=ObjClassCollection(obj_classes))
        return self.items_count > 0

    def to_supervisely(
        self,
        item,
        meta: ProjectMeta,
        renamed_classes: dict = {},
        renamed_tags: dict = {},
        static_transformations: StaticTransformations = None,
    ) -> PointcloudEpisodeAnnotation:
        static_transformations.set_cam2world(item._poses_path)

        frame_cnt = len(item._frame_paths)
        objs, frames = [], []

        frame_idx_to_figures = {idx: [] for idx in range(frame_cnt)}
        for obj in item._ann_data.get_objects():
            pcd_obj = PointcloudEpisodeObject(meta.get_obj_class(obj.name))
            objs.append(pcd_obj)

            for idx in range(frame_cnt):
                if obj.start_frame <= idx <= obj.end_frame:
                    tr_matrix = static_transformations.world_to_velo_transformation(obj, idx)
                    geom = convert_kitti_cuboid_to_supervisely_geometry(tr_matrix)
                    frame_idx_to_figures[idx].append(PointcloudFigure(pcd_obj, geom, idx))
        for idx, figures in frame_idx_to_figures.items():
            frame = PointcloudEpisodeFrame(idx, figures)
            frames.append(frame)
        obj_collection = PointcloudEpisodeObjectCollection(objs)
        frame_collection = PointcloudEpisodeFrameCollection(frames)
        return PointcloudEpisodeAnnotation(
            frame_cnt, objects=obj_collection, frames=frame_collection
        )

    def upload_dataset(self, api: Api, dataset_id: int, batch_size: int = 1, log_progress=True):
        meta, renamed_classes, renamed_tags = self.merge_metas_with_conflicts(api, dataset_id)

        dataset_info = api.dataset.get_info_by_id(dataset_id)
        if log_progress:
            progress, progress_cb = self.get_progress(sum([len(item._frame_paths) for item in self._items]), "Converting pointcloud episodes...")
        else:
            progress_cb = None
        static_transformations = StaticTransformations(self._calib_path)
        scene_ds = dataset_info
        multiple_items = self.items_count > 1
        for item in self._items:
            scene_ds = api.dataset.create(dataset_info.project_id, item._scene_name, parent_id=dataset_id) if multiple_items else dataset_info
            frame_to_pcd_ids = {}
            for idx, frame_path in enumerate(item._frame_paths):
                # * Convert pointcloud from ".bin" to ".pcd"
                pcd_path = str(Path(frame_path).with_suffix(".pcd"))
                if file_exists(pcd_path):
                    logger.warning(f"Overwriting file with path: {pcd_path}")
                convert_bin_to_pcd(frame_path, pcd_path)

                # * Upload pointcloud
                pcd_name = get_file_name_with_ext(pcd_path)
                info = api.pointcloud_episode.upload_path(scene_ds.id, pcd_name, pcd_path, {"frame": idx})
                pcd_id = info.id
                frame_to_pcd_ids[idx] = pcd_id

                # * Clean up
                silent_remove(pcd_path)

                if log_progress:
                    progress_cb(1)

            # * Upload photocontext
            rimage_jsons = []
            cam_names = []
            hashes = api.pointcloud_episode.upload_related_images(
                [rimage_path for _, rimage_path in item._related_images]
            )
            for (cam_name, rimage_path), img, pcd_id in zip(
                item._related_images, hashes, list(frame_to_pcd_ids.values())
            ):
                cam_num = int(cam_name[-1])
                rimage_info = convert_calib_to_image_meta(
                    get_file_name(rimage_path), static_transformations, cam_num
                )
                image_json = {
                    ApiField.ENTITY_ID: pcd_id,
                    ApiField.NAME: cam_name,
                    ApiField.HASH: img,
                    ApiField.META: rimage_info[ApiField.META],
                }
                rimage_jsons.append(image_json)
                cam_names.append(cam_name)
            if rimage_jsons:
                api.pointcloud_episode.add_related_images(rimage_jsons, cam_names)

            # * Convert annotation and upload
            try:
                ann = self.to_supervisely(
                    item, meta, renamed_classes, renamed_tags, static_transformations
                )
                api.pointcloud_episode.annotation.append(scene_ds.id, ann, frame_to_pcd_ids)
            except Exception as e:
                logger.error(
                    f"Failed to upload annotation for scene: {scene_ds.name}. Error: {repr(e)}",
                    stack_info=False,
                )
                continue

            logger.info(f"Dataset ID:{scene_ds.id} has been successfully uploaded.")

        if log_progress:
            if is_development():
                progress.close()



import copy
from typing import Any, Dict, List, Optional, Union

import numpy as np
import pandas as pd

from supervisely.app import DataJson
from supervisely.app.widgets import Widget


class PackerUnpacker:
    SUPPORTED_TYPES = tuple([dict, pd.DataFrame])

    @staticmethod
    def validate_sizes(unpacked_data):
        for row in unpacked_data["data"]:
            if len(row) != len(unpacked_data["columns"]):
                raise ValueError(
                    "Sizes mismatch:\n"
                    f'{len(row)} != {len(unpacked_data["columns"])}\n'
                    f"{row}\n"
                    f'{unpacked_data["columns"]}'
                )

    @staticmethod
    def unpack_data(data, unpacker_cb):
        unpacked_data = unpacker_cb(data)
        PackerUnpacker.validate_sizes(unpacked_data)
        return unpacked_data

    @staticmethod
    def pack_data(data, packer_cb):
        packed_data = packer_cb(data)
        return packed_data

    @staticmethod
    def dict_unpacker(data: dict):
        unpacked_data = {"columns": data["columns"], "data": data["data"]}

        return unpacked_data

    @staticmethod
    def pandas_unpacker(data: pd.DataFrame):
        data = data.where(pd.notnull(data), None)
        data = data.astype(object).replace(np.nan, "-")  # may be None

        unpacked_data = {
            "columns": data.columns.to_list(),
            "data": data.values.tolist(),
        }
        return unpacked_data

    @staticmethod
    def dict_packer(data):
        packed_data = {"columns": data["columns"], "data": data["data"]}
        return packed_data

    @staticmethod
    def pandas_packer(data):
        packed_data = pd.DataFrame(data=data["data"], columns=data["columns"])
        return packed_data


DATATYPE_TO_PACKER = {
    pd.DataFrame: PackerUnpacker.pandas_packer,
    dict: PackerUnpacker.dict_packer,
}

DATATYPE_TO_UNPACKER = {
    pd.DataFrame: PackerUnpacker.pandas_unpacker,
    dict: PackerUnpacker.dict_unpacker,
}


class ClassicTable(Widget):
    """ClassicTable is a widget in Supervisely that is used for displaying and manipulating data in a table format.
    It is similar to the Table widget but with fewer customization options and functionalities.

    Read about it in `Developer Portal <https://developer.supervisely.com/app-development/widgets/tables/classictable>`_
        (including screenshots and examples).

    :param data: Data of table in different formats, see Usage examples.
    :type data: Optional[Union[pd.DataFrame, Dict[str, List[Any]]]]
    :param columns: List of column names.
    :type columns: Optional[List]
    :param fixed_columns_num: Number of fixed columns.
    :type fixed_columns_num: Optional[int]
    :param widget_id: The ID of the widget.
    :type widget_id: Optional[str]

    :Usage example:
    :Usage example:
    .. code-block:: python

        from supervisely.app.widgets import ClassicTable

        # Usage example 1: Pandas DataFrame
        table_data = pd.DataFrame(
            data=[
                ["row_1_column_1", "row_1_column_2", "row_1_column_3"],
                ["row_2_column_1", "row_2_column_2", "row_2_column_3"],
                ["row_3_column_1", "row_3_column_2", "row_3_column_3"],
            ], columns=["col_name_1", "col_name_2", "col_name_3"]
        )

        classic_table = ClassicTable(data=table_data)

        # Usage example 2: Python dict
        table_data =  {
            "columns_names": ["col_name_1", "col_name_2"],
            "values_by_rows": [
        ["row_1_column_1", "row_1_column_2"],
        ["row_2_column_1", "row_2_column_2"],
            ]
        }

        columns = ["col_name_1", "col_name_2"]

        classic_table = ClassicTable(data=table_data, columns=columns)

    """

    class Routes:
        CELL_CLICKED = "cell_clicked_cb"

    def __init__(
        self,
        data: Optional[Union[pd.DataFrame, Dict[str, List[Any]]]] = None,
        columns: Optional[List] = None,
        fixed_columns_num: Optional[int] = None,
        widget_id: Optional[str] = None,
    ):
        self._supported_types = PackerUnpacker.SUPPORTED_TYPES

        self._parsed_data = None
        self._data_type = None

        self._update_table_data(input_data=pd.DataFrame(data=data, columns=columns))

        self._fix_columns = fixed_columns_num

        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self) -> Dict[str, Any]:
        """Returns dictionary with widget data, which defines the appearance and behavior of the widget.

        Dictionary contains the following fields:
            - table_data: Data of table in different
            - table_options: dictionary with table options
                - fixColumns: Number of fixed columns.
        """
        return {
            "table_data": self._parsed_data,
            "table_options": {
                "fixColumns": self._fix_columns,
            },
        }

    def get_json_state(self) -> Dict[str, Dict]:
        """Returns dictionary with widget state.

        Dictionary contains the following fields:
            - selected_row: dictionary with empty dictionary
        """
        return {"selected_row": {}}

    def _update_table_data(self, input_data):
        if input_data is not None:
            self._parsed_data = copy.deepcopy(self._get_unpacked_data(input_data=input_data))
        else:
            self._parsed_data = {"columns": [], "data": []}
            self._data_type = dict

    def _get_packed_data(self, input_data, data_type):
        return PackerUnpacker.pack_data(data=input_data, packer_cb=DATATYPE_TO_PACKER[data_type])

    def _get_unpacked_data(self, input_data):
        input_data_type = type(input_data)

        if input_data_type not in self._supported_types:
            raise TypeError(
                f"Cannot parse input data, please use one of supported datatypes: {self._supported_types}\n"
                """
                            1. Pandas Dataframe \n
                            2. Python dict with structure {
                                        'columns': ['col_name_1', 'col_name_2', ...],
                                        'data': [
                                                            ['row_1_column_1', 'row_1_column_2', ...],
                                                            ['row_2_column_1', 'row_2_column_2', ...],
                                                            ...
                                                          ]
                                      }
                            """
            )

        return PackerUnpacker.unpack_data(
            data=input_data, unpacker_cb=DATATYPE_TO_UNPACKER[input_data_type]
        )

    @property
    def fixed_columns_num(self) -> int:
        """Returns number of fixed columns.

        :return: Number of fixed columns.
        :rtype: int
        """
        return self._fix_columns

    @fixed_columns_num.setter
    def fixed_columns_num(self, value: int) -> None:
        """Sets number of fixed columns.

        :param value: Number of fixed columns.
        :type value: int
        """
        self._fix_columns = value
        DataJson()[self.widget_id]["table_options"]["fixColumns"] = self._fix_columns

    def to_json(self) -> Dict[str, Any]:
        """Returns table data in JSON format.

        :return: Table data in JSON format.
        :rtype: Dict[str, Any]
        """
        return self._get_packed_data(self._parsed_data, dict)

    def to_pandas(self) -> pd.DataFrame:
        """Returns table data in pandas DataFrame format.

        :return: Table data in pandas DataFrame format.
        :rtype: pd.DataFrame
        """
        return self._get_packed_data(self._parsed_data, pd.DataFrame)

    def read_json(self, value: Dict[str, Any]) -> None:
        """Reads table data in JSON format.

        :param value: Table data in JSON format.
        :type value: Dict[str, Any]
        """
        self._update_table_data(input_data=value)
        DataJson()[self.widget_id]["table_data"] = self._parsed_data

    def read_pandas(self, value: pd.DataFrame) -> None:
        """Reads table data in pandas DataFrame format.

        :param value: Table data in pandas DataFrame format.
        :type value: pd.DataFrame
        """
        self._update_table_data(input_data=value)
        DataJson()[self.widget_id]["table_data"] = self._parsed_data

    def insert_row(self, data: List[Any], index: Optional[int] = -1) -> None:
        """Inserts row to the table to the specified index.

        :param data: List of values for new row.
        :type data: List[Any]
        :param index: Index of new row, defaults to -1.
        :type index: Optional[int], optional
        """
        PackerUnpacker.validate_sizes({"columns": self._parsed_data["columns"], "data": [data]})

        table_data = self._parsed_data["data"]
        index = len(table_data) if index > len(table_data) or index < 0 else index

        self._parsed_data["data"].insert(index, data)
        DataJson()[self.widget_id]["table_data"] = self._parsed_data

    def pop_row(self, index: Optional[int] = -1) -> List[Any]:
        """Removes row from the table from the specified index and returns it.

        :param index: Index of row to remove, defaults to -1.
        :type index: Optional[int], optional
        :return: Removed row.
        :rtype: List[Any]
        """
        index = (
            len(self._parsed_data["data"]) - 1
            if index > len(self._parsed_data["data"]) or index < 0
            else index
        )

        if len(self._parsed_data["data"]) != 0:
            popped_row = self._parsed_data["data"].pop(index)
            DataJson()[self.widget_id]["table_data"] = self._parsed_data
            return popped_row

    def get_selected_cell(self, state: Dict[str, Dict]) -> Dict[str, Any]:
        """Returns dictionary with information about selected cell.

        Dictionary contains the following fields:
            - row_index: Index of selected row.
            - col_index: Index of selected column.
            - row_data: Dictionary with data of selected row.
            - cell_data: Tuple with data of selected cell.

        :param state: Dictionary with widget state.
        :type state: Dict[str, Dict]
        :return: Dictionary with information about selected cell.
        :rtype: Dict[str, Any]
        """
        row_index = state[self.widget_id]["selected_row"].get("selectedRow")
        col_index = state[self.widget_id]["selected_row"].get("selectedColumn")
        row_data = state[self.widget_id]["selected_row"].get("selectedRowData", {})

        return {
            "row_index": row_index,
            "col_index": col_index,
            "row_data": row_data,
            "cell_data": list(row_data.items())[int(col_index)]
            if col_index is not None and row_data is not None
            else None,
        }



import lap  # pylint: disable=import-error
import numpy as np
from cython_bbox import bbox_overlaps as bbox_ious  # pylint: disable=import-error
from scipy.spatial.distance import cdist

from supervisely.nn.tracker.utils import kalman_filter


def linear_assignment(cost_matrix, thresh):
    if cost_matrix.size == 0:
        return (
            np.empty((0, 2), dtype=int),
            tuple(range(cost_matrix.shape[0])),
            tuple(range(cost_matrix.shape[1])),
        )
    matches, unmatched_a, unmatched_b = [], [], []
    cost, x, y = lap.lapjv(cost_matrix, extend_cost=True, cost_limit=thresh)
    for ix, mx in enumerate(x):
        if mx >= 0:
            matches.append([ix, mx])
    unmatched_a = np.where(x < 0)[0]
    unmatched_b = np.where(y < 0)[0]
    matches = np.asarray(matches)
    return matches, unmatched_a, unmatched_b


def ious(atlbrs, btlbrs):
    """
    Compute cost based on IoU
    :type atlbrs: list[tlbr] | np.ndarray
    :type atlbrs: list[tlbr] | np.ndarray

    :rtype ious np.ndarray
    """
    ious = np.zeros((len(atlbrs), len(btlbrs)), dtype=float)
    if ious.size == 0:
        return ious

    ious = bbox_ious(
        np.ascontiguousarray(atlbrs, dtype=float), np.ascontiguousarray(btlbrs, dtype=float)
    )

    return ious


def iou_distance(atracks, btracks):
    """
    Compute cost based on IoU
    :type atracks: list[STrack]
    :type btracks: list[STrack]

    :rtype cost_matrix np.ndarray
    """

    if (len(atracks) > 0 and isinstance(atracks[0], np.ndarray)) or (
        len(btracks) > 0 and isinstance(btracks[0], np.ndarray)
    ):
        atlbrs = atracks
        btlbrs = btracks
    else:
        atlbrs = [track.tlbr for track in atracks]
        btlbrs = [track.tlbr for track in btracks]
    _ious = ious(atlbrs, btlbrs)
    cost_matrix = 1 - _ious

    return cost_matrix


def embedding_distance(tracks, detections, metric="cosine"):
    """
    :param tracks: list[STrack]
    :param detections: list[BaseTrack]
    :param metric:
    :return: cost_matrix np.ndarray
    """

    cost_matrix = np.zeros((len(tracks), len(detections)), dtype=float)
    if cost_matrix.size == 0:
        return cost_matrix
    det_features = np.asarray([track.curr_feat for track in detections], dtype=float)
    track_features = np.asarray([track.smooth_feat for track in tracks], dtype=float)

    cost_matrix = np.maximum(
        0.0, cdist(track_features, det_features, metric)
    )  # / 2.0  # Nomalized features
    return cost_matrix


def fuse_motion(kf, cost_matrix, tracks, detections, only_position=False, lambda_=0.98):
    if cost_matrix.size == 0:
        return cost_matrix
    gating_dim = 2 if only_position else 4
    gating_threshold = kalman_filter.chi2inv95[gating_dim]
    # measurements = np.asarray([det.to_xyah() for det in detections])
    measurements = np.asarray([det.to_xywh() for det in detections])
    for row, track in enumerate(tracks):
        gating_distance = kf.gating_distance(
            track.mean, track.covariance, measurements, only_position, metric="maha"
        )
        cost_matrix[row, gating_distance > gating_threshold] = np.inf
        cost_matrix[row] = lambda_ * cost_matrix[row] + (1 - lambda_) * gating_distance
    return cost_matrix


def fuse_iou(cost_matrix, tracks, detections):
    if cost_matrix.size == 0:
        return cost_matrix
    reid_sim = 1 - cost_matrix
    iou_dist = iou_distance(tracks, detections)
    iou_sim = 1 - iou_dist
    fuse_sim = reid_sim * (1 + iou_sim) / 2
    det_scores = np.array([det.score for det in detections])
    det_scores = np.expand_dims(det_scores, axis=0).repeat(cost_matrix.shape[0], axis=0)
    # fuse_sim = fuse_sim * (1 + det_scores) / 2
    fuse_cost = 1 - fuse_sim
    return fuse_cost


def fuse_score(cost_matrix, detections):
    if cost_matrix.size == 0:
        return cost_matrix
    iou_sim = 1 - cost_matrix
    det_scores = np.array([det.score for det in detections])
    det_scores = np.expand_dims(det_scores, axis=0).repeat(cost_matrix.shape[0], axis=0)
    fuse_sim = iou_sim * det_scores
    fuse_cost = 1 - fuse_sim
    return fuse_cost



from typing import Any, Callable, Dict, List, Literal, Optional, Union

from supervisely.app import StateJson
from supervisely.app.widgets import Widget

SUPPORTED_TAG_WIDGET_TYPES = ["primary", "gray", "success", "warning", "danger"]

# @TODO: fix problem with close transition id


class ElementTagsList(Widget):
    """ElementTagsList widget in Supervisely is a widget that allows users to display multiple elements tags in the UI.

    Read about it in `Developer Portal <https://developer.supervisely.com/app-development/widgets/text-elements/elementtagslist>`_
        (including screenshots and examples).

    :param tags: List of tags
    :type tags: Optional[List[Tag]]
    :param widget_id: An identifier of the widget.
    :type widget_id: str, optional

    :Usage example:
    .. code-block:: python

            from supervisely.app.widgets import ElementTagsList

            element_tags_list = ElementTagsList(
                tags=[
                    ElementTagsList.Tag(
                        text="Tag",
                        type="primary",
                        hit=True,
                        color="#20a0ff",
                    )
                ]
            )
    """

    class Routes:
        CLOSE = "tag_close_cb"

    class Tag:
        """Represents tag in ElementTagsList widget.

        :param text: Tag text
        :type text: Optional[str]
        :param type: Tag type, one of: primary, gray, success, warning, danger
        :type type: Optional[Literal["primary", "gray", "success", "warning", "danger"]]
        :param hit: If True, tag will be highlighted
        :type hit: Optional[bool]
        :param color: Tag color
        :type color: Optional[str]
        :param closable: If True, tag will be closable
        :type closable: Optional[bool]
        :param close_transition: If True, tag will be closable with transition
        :type close_transition: Optional[bool]
        """

        def __init__(
            self,
            text: str,
            type: Optional[Literal["primary", "gray", "success", "warning", "danger"]] = "primary",
            hit: Optional[bool] = False,
            color: Optional[str] = "",
            closable: Optional[bool] = False,
            close_transition: Optional[bool] = False,
        ):
            self._text = text
            self._type = type
            self._hit = hit
            self._color = color
            self._closable = closable
            self._close_transition = close_transition

        @property
        def text(self) -> str:
            """Returns tag text.

            :return: tag text
            :rtype: str
            """
            return self._text

        @property
        def type(self) -> str:
            """Returns tag type.

            :return: tag type
            :rtype: str
            """
            return self._type

        @property
        def hit(self) -> bool:
            """Returns True if tag is highlighted, False otherwise.

            :return: True if tag is highlighted, False otherwise
            :rtype: bool
            """
            return self._hit

        @property
        def color(self) -> str:
            """Returns tag color.

            :return: tag color
            :rtype: str
            """
            return self._color

        def to_json(self) -> Dict[str, Union[str, bool]]:
            """Returns JSON representation of tag.

            Dictionary contains the following fields:
                - text: Tag text
                - type: Tag type, one of: primary, gray, success, warning, danger
                - hit: If True, tag will be highlighted
                - color: Tag color
                - closable: If True, tag will be closable
                - close_transition: If True, tag will be closable with transition

            :return: dictionary with tag data
            :rtype: Dict[str, Union[str, bool]]
            """
            return {
                "text": self._text,
                "type": self._type,
                "hit": self._hit,
                "color": self._color,
                "closable": self._closable,
                "close_transition": self._close_transition,
            }

        @classmethod
        def from_json(cls, tag_json: Dict[str, Union[str, bool]]) -> "ElementTagsList.Tag":
            """Creates tag from JSON representation.

            :param tag_json: JSON representation of tag
            :type tag_json: Dict[str, Union[str, bool]]
            :return: tag
            :rtype: ElementTagsList.Tag
            """
            return cls(
                tag_json["text"],
                tag_json["type"],
                tag_json["hit"],
                tag_json["color"],
                tag_json["closable"],
                tag_json["close_transition"],
            )

    def __init__(
        self,
        tags: Optional[List[Tag]] = [],
        widget_id: Optional[str] = None,
    ):
        self._clicked_tag = None

        self._validate_tags(tags)
        self._tags = tags

        self._close_handled = False

        super().__init__(widget_id=widget_id, file_path=__file__)

    def _validate_tags(self, tags):
        for tag in tags:
            if not isinstance(tag, ElementTagsList.Tag):
                raise ValueError("Tag type should be class ElementTagsList.Tag")

    def get_json_data(self) -> Dict:
        """ElementTagsList widget has no data, the method returns an empty dict."""
        return {}

    def get_json_state(self) -> Dict[str, List[Dict[str, Union[str, bool]]]]:
        """Returns dictionary with widget state.

        The dictionary contains the following fields:
            - tags: List of JSON representations of tags

        :return: dictionary with widget state
        :rtype: Dict[str, List[Dict[str, Union[str, bool]]]]
        """
        return {"tags": [tag.to_json() for tag in self._tags]}

    def set_tags(self, tags: List[Tag]) -> None:
        """Sets tags for ElementTagsList widget.
        This method overwrites all previous tags.
        To add tags, use add_tags method.

        :param tags: List of tags
        :type tags: List[Tag]]
        """

        self._tags = tags
        self.update_state()
        StateJson().send_changes()

    def get_tags(self) -> List[Tag]:
        """Returns current tags.

        :return: current tags
        :rtype: List[Tag]
        """
        return [ElementTagsList.Tag.from_json(tag) for tag in StateJson()[self.widget_id]["tags"]]

    def add_tags(self, tags: List[Tag]) -> None:
        """Adds tags to ElementTagsList widget.
        This method adds tags to the end of the list.
        To replace all tags, use set_tags method.

        :param tags: List of tags
        :type tags: List[Tag]]
        """
        self._tags = self.get_tags()
        self._tags.extend(tags)
        self.update_state()
        StateJson().send_changes()

    def close(self, func: Callable[[List[Tag]], Any]) -> Callable[[], None]:
        """Decorator for function that will be called when tag is closed.

        :param func: Function that will be called when tag is closed
        :type func: Callable[[List[Tag]], Any]
        :return: Decorated function
        :rtype: Callable[[], None]
        """
        route_path = self.get_route_path(ElementTagsList.Routes.CLOSE)
        server = self._sly_app.get_server()
        self._changes_handled = True

        @server.post(route_path)
        async def _click():
            res = self.get_tags()
            func(res)

        return _click



from typing import Dict, Optional, Union

from supervisely.app import DataJson, StateJson
from supervisely.app.widgets import Widget


class Badge(Widget):
    """Badge widget in Supervisely is a versatile tool for displaying notifications or counts on elements such as buttons, text.
    Read about it in `Developer Portal <https://developer.supervisely.com/app-development/widgets/status-elements/badge>`_
        (including screenshots and examples).

    :param value: Value to be displayed on the badge.
    :type value: Optional[Union[int, str, float]]
    :param widget: Widget to be displayed on the badge.
    :type widget: Optional[Widget]
    :param max: Maximum value of the badge. If value is greater than max, max will be displayed on the badge.
    :type max: Optional[Union[int, float]]
    :param is_dot: If True, the badge will be displayed as a dot.
    :type is_dot: Optional[bool]
    :param hidden: If True, the badge will be hidden.
    :type hidden: Optional[bool]
    :param widget_id: Unique widget identifier.
    :type widget_id: Optional[str]

    :Usage example:
    .. code-block:: python
        from supervisely.app.widgets import Badge

        badge = Badge(value=5, max=10)
    """

    def __init__(
        self,
        value: Optional[Union[int, str, float]] = None,
        widget: Optional[Widget] = None,
        max: Optional[Union[int, float]] = None,
        is_dot: Optional[bool] = False,
        hidden: Optional[bool] = False,
        widget_id: Optional[str] = None,
    ):
        self._value = value
        self._widget = widget
        self._max = max if type(max) in [int, float] else None
        self._hidden = hidden
        self._is_dot = is_dot

        if self._value is None and self._widget is not None:
            self._is_dot = True

        if self._is_dot is True and self._value is None:
            self._value = 0

        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self) -> Dict[str, Union[int, float, bool]]:
        """Returns dictionary with widget data, which defines the appearance and behavior of the widget.

        Dictionary contains the following fields:
            - max: Maximum value of the badge. If value is greater than max, max will be displayed on the badge.
            - isDot: If True, the badge will be displayed as a dot.
            - hidden: If True, the badge will be hidden.

        :return: Dictionary with widget data.
        :rtype: Dict[str, Union[int, float, bool]]
        """
        res = {}
        res["max"] = self._max
        res["isDot"] = self._is_dot
        res["hidden"] = self._hidden
        return res

    def get_json_state(self) -> Dict[str, Union[str, int, float]]:
        """Returns dictionary with widget state.

        Dictionary contains the following fields:
            - value: Value to be displayed on the badge.

        :return: Dictionary with widget state.
        :rtype: Dict[str, Union[str, int, float]]
        """
        return {"value": self._value}

    def set_value(self, value: Union[str, int, float]) -> None:
        """Sets value to be displayed on the badge.

        :param value: Value to be displayed on the badge.
        :type value: Union[str, int, float]
        """
        self._value = value
        StateJson()[self.widget_id]["value"] = self._value
        StateJson().send_changes()

    def get_value(self) -> Union[str, int, float]:
        """Returns value to be displayed on the badge.

        :return: Value to be displayed on the badge.
        :rtype: Union[str, int, float]
        """
        if "value" not in StateJson()[self.widget_id].keys():
            return None
        value = StateJson()[self.widget_id]["value"]
        return value

    def clear(self) -> None:
        """Clears the value of the badge."""
        self._value = None
        StateJson()[self.widget_id]["value"] = self._value
        StateJson().send_changes()

    def hide_badge(self) -> None:
        """Hides the badge."""
        self._hidden = True
        DataJson()[self.widget_id]["hidden"] = self._hidden
        DataJson().send_changes()

    def show_badge(self) -> None:
        """Shows the badge."""
        self._hidden = False
        DataJson()[self.widget_id]["hidden"] = self._hidden
        DataJson().send_changes()

    def toggle_visibility(self) -> None:
        """Toggles the visibility of the badge.
        If the badge is hidden, it will be shown.
        If the badge is shown, it will be hidden.
        """
        self._hidden = not self._hidden
        DataJson()[self.widget_id]["hidden"] = self._hidden
        DataJson().send_changes()



import os
from typing import List, Tuple

import supervisely.convert.image.label_me.label_me_helper as label_me_helper
from supervisely import ProjectMeta, logger
from supervisely.annotation.annotation import Annotation
from supervisely.convert.base_converter import AvailableImageConverters
from supervisely.convert.image.image_converter import ImageConverter
from supervisely.io.fs import get_file_ext, get_file_name
from supervisely.io.json import load_json_file


class LabelmeConverter(ImageConverter):

    def __str__(self):
        return AvailableImageConverters.LABEL_ME

    @property
    def ann_ext(self) -> str:
        return ".json"

    def validate_ann_file(self, ann_path: str) -> bool:
        """Validate LabelMe annotation file."""

        ann_json = load_json_file(ann_path)
        if not isinstance(ann_json, dict):
            return False
        shapes = ann_json.get("shapes")
        if shapes is None or not isinstance(shapes, list):
            return False
        if ann_json.get("imagePath") is None and ann_json.get("imageData") is None:
            return False
        return True

    def validate_format(self) -> bool:
        logger.debug(f"Validating format: {self.__str__()}")
        items, meta = self._find_images_with_annotations()
        if len(items) == 0:
            logger.debug(f"No LabelMe data found in {self._input_data}.")
            return False
        else:
            self._items = items
            self._meta = meta
            logger.debug(f"Found {self.items_count} LabelMe items.")
            return True

    def _find_images_with_annotations(self) -> Tuple[List[ImageConverter.Item], ProjectMeta]:
        items = []
        images, anns = {}, {}
        meta = ProjectMeta()
        for root, _, files in os.walk(self._input_data):
            for file in files:
                file_name_no_ext = get_file_name(file)
                full_path = os.path.join(root, file)
                ext = get_file_ext(full_path)
                if ext == self.ann_ext:
                    if self.validate_ann_file(full_path):
                        meta = label_me_helper.update_meta_from_labelme_annotation(meta, full_path)
                        anns[file_name_no_ext] = full_path
                elif self.is_image(full_path):
                    images[file_name_no_ext] = full_path
        for name_noext, ann_path in anns.items():
            image_path = label_me_helper.get_image_from_data(ann_path, images.get(name_noext))
            if image_path is not None:
                item = self.Item(image_path, ann_path)
                items.append(item)
        return items, meta

    def to_supervisely(
        self,
        item: ImageConverter.Item,
        meta: ProjectMeta = None,
        renamed_classes: dict = None,
        renamed_tags: dict = None,
    ) -> Annotation:
        """Convert to Supervisely format."""

        item.set_shape()
        ann = label_me_helper.create_supervisely_annotation(
            item,
            meta,
            renamed_classes,
        )
        return ann



# coding: utf-8
"""single prediction object"""

from __future__ import annotations

import atexit
import os
import tempfile
from os import PathLike
from pathlib import Path
from typing import Dict, List, Optional, Union

import numpy as np
import requests

from supervisely._utils import get_valid_kwargs, logger, rand_str
from supervisely.annotation.annotation import Annotation
from supervisely.annotation.label import Label
from supervisely.annotation.tag import Tag
from supervisely.annotation.tag_meta import TagValueType
from supervisely.api.api import Api
from supervisely.convert.image.sly.sly_image_helper import get_meta_from_annotation
from supervisely.geometry.bitmap import Bitmap
from supervisely.geometry.rectangle import Rectangle
from supervisely.imaging.image import read as read_image
from supervisely.imaging.image import read_bytes as read_image_bytes
from supervisely.imaging.image import write as write_image
from supervisely.io.fs import (
    clean_dir,
    dir_empty,
    ensure_base_path,
    get_file_ext,
    mkdir,
)
from supervisely.project.project_meta import ProjectMeta
from supervisely.video.video import VideoFrameReader


class Prediction:
    _temp_dir = os.path.join(tempfile.gettempdir(), "prediction_files")
    __cleanup_registered = False

    def __init__(
        self,
        annotation_json: Dict,
        source: Union[str, int] = None,
        model_meta: Optional[Union[ProjectMeta, Dict]] = None,
        name: Optional[str] = None,
        path: Optional[str] = None,
        url: Optional[str] = None,
        project_id: Optional[int] = None,
        dataset_id: Optional[int] = None,
        image_id: Optional[int] = None,
        video_id: Optional[int] = None,
        frame_index: Optional[int] = None,
        api: Optional["Api"] = None,
        **kwargs,
    ):
        self.source = source
        if isinstance(annotation_json, Annotation):
            annotation_json = annotation_json.to_json()
        self.annotation_json = annotation_json
        self.model_meta = model_meta
        if isinstance(self.model_meta, dict):
            self.model_meta = ProjectMeta.from_json(self.model_meta)

        self.name = name
        self.path = path
        self.url = url
        self.project_id = project_id
        self.dataset_id = dataset_id
        self.image_id = image_id
        self.video_id = video_id
        self.frame_index = frame_index
        self.extra_data = {}
        if kwargs:
            self.extra_data.update(kwargs)
        self.api = api

        self._annotation = None
        self._boxes = None
        self._masks = None
        self._classes = None
        self._scores = None

        if self.path is None and isinstance(self.source, (str, PathLike)):
            self.path = str(self.source)

    def _init_geometries(self):

        def _get_confidence(label: Label):
            for tag_name in ["confidence", "conf", "score"]:
                conf_tag: Tag = label.tags.get(tag_name, None)
                if conf_tag is not None and conf_tag.meta.value_type != str(
                    TagValueType.ANY_NUMBER
                ):
                    conf_tag = None
                if conf_tag is not None:
                    return conf_tag.value
            return 1

        self._boxes = []
        self._masks = []
        self._classes = []
        self._scores = []
        for label in self.annotation.labels:
            self._classes.append(label.obj_class.name)
            self._scores.append(_get_confidence(label))
            if isinstance(label.geometry, Rectangle):
                rect = label.geometry
            else:
                rect = label.geometry.to_bbox()
            if isinstance(label.geometry, Bitmap):
                self._masks.append(label.geometry.get_mask(self.annotation.img_size))

            self._boxes.append(
                np.array(
                    [
                        rect.top,
                        rect.left,
                        rect.bottom,
                        rect.right,
                    ]
                )
            )
        self._boxes = np.array(self._boxes)
        self._masks = np.array(self._masks)

    @property
    def boxes(self):
        if self._boxes is None:
            self._init_geometries()
        return self._boxes

    @property
    def masks(self):
        if self._masks is None:
            self._init_geometries()
        return self._masks

    @property
    def classes(self):
        if self._classes is None:
            self._init_geometries()
        return self._classes

    @property
    def scores(self):
        if self._scores is None:
            self._init_geometries()
        return self._scores

    @property
    def annotation(self) -> Annotation:
        if self._annotation is None:
            if self.model_meta is None:
                raise ValueError("Model meta is not provided. Cannot create annotation.")
            model_meta = get_meta_from_annotation(self.annotation_json, self.model_meta)
            self._annotation = Annotation.from_json(self.annotation_json, model_meta)
        return self._annotation

    @annotation.setter
    def annotation(self, annotation: Union[Annotation, Dict]):
        if isinstance(annotation, Annotation):
            self._annotation = annotation
            self.annotation_json = annotation.to_json()
        elif isinstance(annotation, dict):
            self._annotation = None
            self.annotation_json = annotation
        else:
            raise ValueError("Annotation must be either a dict or an Annotation object.")

    @property
    def class_idxs(self) -> np.ndarray:
        if self.model_meta is None:
            raise ValueError("Model meta is not provided. Cannot create class indexes.")
        cls_name_to_idx = {
            obj_class.name: i for i, obj_class in enumerate(self.model_meta.obj_classes)
        }
        return np.array([cls_name_to_idx[class_name] for class_name in self.classes])

    @classmethod
    def from_json(cls, json_data: Dict, **kwargs) -> "Prediction":
        kwargs = {**json_data, **kwargs}
        if "annotation_json" in kwargs:
            annotation_json = kwargs.pop("annotation_json")
        elif "annotation" in kwargs:
            annotation_json = kwargs.pop("annotation")
        else:
            raise ValueError("Annotation JSON is required.")
        kwargs = get_valid_kwargs(
            kwargs,
            Prediction.__init__,
            exclude=["self", "annotation_json"],
        )
        return cls(annotation_json, **kwargs)

    def to_json(self):
        return {
            "source": self.source,
            "annotation": self.annotation_json,
            "name": self.name,
            "path": self.path,
            "url": self.url,
            "project_id": self.project_id,
            "dataset_id": self.dataset_id,
            "image_id": self.image_id,
            "video_id": self.video_id,
            "frame_index": self.frame_index,
            **self.extra_data,
        }

    def _clear_temp_files(self):
        if not dir_empty(self._temp_dir):
            clean_dir(self._temp_dir)

    def load_image(self) -> np.ndarray:
        api = self.api
        if self.frame_index is None:
            if self.path is not None:
                return read_image(self.path)
            if self.url is not None:
                ext = get_file_ext(self.url)
                if ext == "":
                    ext = ".jpg"
                r = requests.get(self.url, stream=True, timeout=60)
                r.raise_for_status()
                return read_image_bytes(r.content)
            if self.image_id is not None:
                try:
                    if api is None:
                        api = Api()
                    return api.image.download_np(self.image_id)
                except Exception as e:
                    raise RuntimeError("Failed to load image by ID") from e
            raise ValueError("Cannot load image. No path, URL, image ID, or frame_index provided.")
        if self.video_id is not None:
            if self.frame_index is None:
                raise ValueError("Frame index is not provided for video.")
            try:
                if api is None:
                    api = Api()
                return api.video.frame.download_np(self.video_id, self.frame_index)
            except Exception as e:
                raise RuntimeError("Failed to load frame by video ID") from e
        if self.path is not None:
            return next(VideoFrameReader(self.path, frame_indexes=[self.frame_index]))
        if self.url is not None:
            video_name = Path(self.url).name
            video_path = Path(self._temp_dir) / video_name
            mkdir(self._temp_dir)
            if not video_path.exists():
                with requests.get(self.url, stream=True, timeout=10 * 60) as r:
                    r.raise_for_status()
                    with open(video_path, "wb") as f:
                        for chunk in r.iter_content(chunk_size=8192):
                            f.write(chunk)
            return next(VideoFrameReader(video_path, frame_indexes=[self.frame_index]))
        raise ValueError("Cannot load frame. No path, URL or video ID provided.")

    def visualize(
        self,
        save_path: Optional[str] = None,
        save_dir: Optional[str] = None,
        color: Optional[List[int]] = None,
        thickness: Optional[int] = None,
        opacity: Optional[float] = 0.5,
        draw_tags: Optional[bool] = False,
        fill_rectangles: Optional[bool] = True,
    ) -> np.ndarray:
        if save_dir is not None and save_path is not None:
            raise ValueError("Only one of save_path or save_dir can be provided.")

        mkdir(self._temp_dir)
        if not Prediction.__cleanup_registered:
            atexit.register(self._clear_temp_files)
            Prediction.__cleanup_registered = True

        img = self.load_image()
        self.annotation.draw_pretty(
            bitmap=img,
            color=color,
            thickness=thickness,
            opacity=opacity,
            draw_tags=draw_tags,
            output_path=None,
            fill_rectangles=fill_rectangles,
        )
        if save_dir is not None:
            save_path = save_dir
        if save_path is None:
            return img
        if Path(save_path).suffix == "":
            # is a directory
            if self.name is not None:
                name = self.name
                if self.frame_index is not None:
                    name = f"{name}_{self.frame_index}"
            elif self.image_id is not None:
                name = str(self.image_id)
            elif self.video_id is not None:
                name = str(self.video_id)
            elif self.path is not None:
                name = Path(self.path).name
                if self.frame_index is not None:
                    name = f"{name}_{self.frame_index}"
            else:
                name = f"{rand_str(6)}"
            name = f"vis_{name}.png"
            save_path = os.path.join(save_path, name)
        ensure_base_path(save_path)
        write_image(save_path, img)
        logger.info("Visualization for prediction saved to %s", save_path)
        return img



from typing import Dict, List

from jinja2 import Template

from supervisely.nn.benchmark.visualization.widgets.widget import BaseWidget


class SidebarWidget(BaseWidget):

    def __init__(self, widgets: List[BaseWidget], anchors: List[str]) -> None:
        self.widgets = widgets
        self.anchors = anchors

    def sidebar_template_str(self, widget: BaseWidget):
        button_style = f"{{fontWeight: data.scrollIntoView === '{widget.id}' ? 'bold' : 'normal'}}"

        return f"""
            <div>
                <el-button type="text" @click="data.scrollIntoView='{widget.id}'" :style="{button_style}">
                {widget.title}
                </el-button>
            </div>
        """

    @property
    def html_str(self):
        anchored_widgets = sorted(
            [widget for widget in self.widgets if widget.id in self.anchors],
            key=lambda x: self.anchors.index(x.id),
        )

        sidebar_content = "".join(self.sidebar_template_str(widget) for widget in anchored_widgets)
        main_content = "".join(
            f"""
            <div style="margin-top: 20px;">
                {widget.to_html()}
            </div>
            """
            for widget in self.widgets
        )
        template = f"""
            <sly-iw-sidebar
                :options="{{ height: 'calc(100vh - 130px)', clearMainPanelPaddings: true, leftSided: false, disableResize: true, sidebarWidth: 300 }}"
            >
                <div slot="sidebar">
                    {sidebar_content}
                </div>
                <div style="padding-right: 35px;">
                    {main_content}
                </div>
            </sly-iw-sidebar>
        """

        return template

    def to_html(self) -> str:
        return Template(self.html_str).render()

    def save_data(self, basepath: str) -> None:
        for widget in self.widgets:
            widget.save_data(basepath)

    def get_state(self) -> Dict:
        state = {}
        for widget in self.widgets:
            state.update(widget.get_state())
        return state



# from supervisely.app.v1.constants import *
# import supervisely.app.v1.widgets

# server-address/apps/designer
# <!--    <sly-notification content="<Notification text in markdown>" :options="{ type: 'note|info|warning|error', name: '<notification header>'}" />-->
# {
# 	"a": "...",
# 	"context_menu": {
# 			"target": ["images_project", "images_dataset", "labeling_job", "team", "workspace"]
#                          "DELME show_description (false by default)": true,
# 			"context_root (optional)": "Download as / Run App (default) / Report",
#                          "context_category (optional)": "Objects"
# 		}
# }

# context_root: Download as, Run App (default), Report
# [![Views](https://app.supervisely.com/public/api/v3/ecosystem.counters?repo=supervisely-ecosystem/roads-test&counter=views&label=custom)](https://supervisely.com)



import os
from pathlib import Path

from supervisely import PointcloudAnnotation, ProjectMeta, is_development, logger
from supervisely.api.api import Api, ApiField
from supervisely.convert.base_converter import AvailablePointcloudConverters
from supervisely.convert.pointcloud.kitti_3d import kitti_3d_helper
from supervisely.convert.pointcloud.pointcloud_converter import PointcloudConverter
from supervisely.io.fs import (
    dirs_filter,
    file_exists,
    get_file_ext,
    get_file_name,
    get_file_name_with_ext,
    list_files,
    silent_remove,
)
from supervisely.pointcloud_annotation.pointcloud_object_collection import (
    PointcloudObjectCollection,
)


class KITTI3DConverter(PointcloudConverter):
    def __str__(self) -> str:
        return AvailablePointcloudConverters.KITTI3D

    @property
    def key_file_ext(self) -> str:
        return ".bin"

    def validate_format(self) -> bool:
        def _file_filter_fn(file_path):
            return get_file_ext(file_path).lower() == self.key_file_ext

        def _dir_filter_fn(path):
            return all([(Path(path) / name).exists() for name in kitti_3d_helper.FOLDER_NAMES])

        input_paths = [d for d in dirs_filter(self._input_data, _dir_filter_fn)]
        if len(input_paths) == 0:
            return False

        input_path = input_paths[0]
        velodyne_dir = os.path.join(input_path, "velodyne")
        image_2_dir = os.path.join(input_path, "image_2")
        label_2_dir = os.path.join(input_path, "label_2")
        calib_dir = os.path.join(input_path, "calib")

        self._items = []
        velodyne_files = list_files(velodyne_dir, filter_fn=_file_filter_fn)
        if len(velodyne_files) == 0:
            return False

        kitti_labels = []
        for velodyne_path in velodyne_files:
            file_name = get_file_name(velodyne_path)
            image_path = os.path.join(image_2_dir, f"{file_name}.png")
            label_path = os.path.join(label_2_dir, f"{file_name}.txt")
            calib_path = os.path.join(calib_dir, f"{file_name}.txt")
            if not file_exists(image_path):
                logger.debug(f"Skipping item: {velodyne_path}. Image not found.")
                continue
            if not file_exists(label_path):
                logger.debug(f"Skipping item: {velodyne_path}. Label not found.")
                continue
            if not file_exists(calib_path):
                logger.debug(f"Skipping item: {velodyne_path}. Calibration not found.")
                continue

            label = kitti_3d_helper.read_kitti_label(label_path, calib_path)
            kitti_labels.append(label)
            self._items.append(self.Item(velodyne_path, label, (image_path, calib_path)))

        self._meta = kitti_3d_helper.convert_labels_to_meta(kitti_labels)
        return self.items_count > 0

    def to_supervisely(
        self,
        item: PointcloudConverter.Item,
        meta: ProjectMeta,
        renamed_classes: dict = {},
        renamed_tags: dict = {},
    ) -> PointcloudAnnotation:
        label = item.ann_data
        objs, figures = kitti_3d_helper.convert_label_to_annotation(label, meta, renamed_classes)
        return PointcloudAnnotation(PointcloudObjectCollection(objs), figures)

    def upload_dataset(self, api: Api, dataset_id: int, batch_size: int = 1, log_progress=True):
        meta, renamed_classes, renamed_tags = self.merge_metas_with_conflicts(api, dataset_id)

        if log_progress:
            progress, progress_cb = self.get_progress(self.items_count, "Converting pointclouds...")
        else:
            progress_cb = None

        for item in self._items:
            # * Convert pointcloud from ".bin" to ".pcd"
            pcd_path = str(Path(item.path).with_suffix(".pcd"))
            if file_exists(pcd_path):
                logger.warning(f"Overwriting file with path: {pcd_path}")
            kitti_3d_helper.convert_bin_to_pcd(item.path, pcd_path)

            # * Upload pointcloud
            pcd_name = get_file_name_with_ext(pcd_path)
            info = api.pointcloud.upload_path(dataset_id, pcd_name, pcd_path, {})
            pcd_id = info.id

            # * Convert annotation and upload
            ann = self.to_supervisely(item, meta, renamed_classes, renamed_tags)
            api.pointcloud.annotation.append(pcd_id, ann)

            # * Upload related images
            image_path, calib_path = item._related_images
            rimage_info = kitti_3d_helper.convert_calib_to_image_meta(image_path, calib_path)

            image_jsons = []
            camera_names = []
            img = api.pointcloud.upload_related_image(image_path)
            image_jsons.append(
                {
                    ApiField.ENTITY_ID: pcd_id,
                    ApiField.NAME: get_file_name_with_ext(rimage_info[ApiField.NAME]),
                    ApiField.HASH: img,
                    ApiField.META: rimage_info[ApiField.META],
                }
            )
            camera_names.append(rimage_info[ApiField.META]["deviceId"])
            if len(image_jsons) > 0:
                api.pointcloud.add_related_images(image_jsons, camera_names)

            # * Clean up
            silent_remove(pcd_path)
            if log_progress:
                progress_cb(1)

        logger.info(f"Dataset ID:{dataset_id} has been successfully uploaded.")

        if log_progress:
            if is_development():
                progress.close()



import cv2
import matplotlib.pyplot as plt
import numpy as np
import torch  # pylint: disable=import-error
import torch.nn.functional as F  # pylint: disable=import-error
from fastreid.config import get_cfg  # pylint: disable=import-error
from fastreid.modeling.meta_arch import build_model  # pylint: disable=import-error
from fastreid.utils.checkpoint import Checkpointer  # pylint: disable=import-error

# from torch.backends import cudnn


# cudnn.benchmark = True


def setup_cfg(config_file, opts):
    # load config from file and command-line arguments
    cfg = get_cfg()
    cfg.merge_from_file(config_file)
    cfg.merge_from_list(opts)
    cfg.MODEL.BACKBONE.PRETRAIN = False

    cfg.freeze()

    return cfg


def postprocess(features):
    # Normalize feature to compute cosine distance
    features = F.normalize(features)
    features = features.cpu().data.numpy()
    return features


def preprocess(image, input_size):
    if len(image.shape) == 3:
        padded_img = np.ones((input_size[1], input_size[0], 3), dtype=np.uint8) * 114
    else:
        padded_img = np.ones(input_size) * 114
    img = np.array(image)
    r = min(input_size[1] / img.shape[0], input_size[0] / img.shape[1])
    resized_img = cv2.resize(
        img,
        (int(img.shape[1] * r), int(img.shape[0] * r)),
        interpolation=cv2.INTER_LINEAR,
    )
    padded_img[: int(img.shape[0] * r), : int(img.shape[1] * r)] = resized_img

    return padded_img, r


class FastReIDInterface:
    def __init__(self, config_file, weights_path, device, batch_size=8):
        super(FastReIDInterface, self).__init__()
        if device != "cpu":
            self.device = "cuda"
        else:
            self.device = "cpu"

        self.batch_size = batch_size

        self.cfg = setup_cfg(config_file, ["MODEL.WEIGHTS", weights_path])

        self.model = build_model(self.cfg)
        self.model.eval()

        Checkpointer(self.model).load(weights_path)

        if self.device != "cpu":
            self.model = self.model.eval().to(device="cuda").half()
        else:
            self.model = self.model.eval()

        self.pH, self.pW = self.cfg.INPUT.SIZE_TEST

    def inference(self, image, detections):

        if detections is None or np.size(detections) == 0:
            return []

        H, W, _ = np.shape(image)

        batch_patches = []
        patches = []
        for d in range(np.size(detections, 0)):
            tlbr = detections[d, :4].astype(np.int_)
            tlbr[0] = max(0, tlbr[0])
            tlbr[1] = max(0, tlbr[1])
            tlbr[2] = min(W - 1, tlbr[2])
            tlbr[3] = min(H - 1, tlbr[3])
            patch = image[tlbr[1] : tlbr[3], tlbr[0] : tlbr[2], :]

            # the model expects RGB inputs
            patch = patch[:, :, ::-1]

            # Apply pre-processing to image.
            patch = cv2.resize(
                patch, tuple(self.cfg.INPUT.SIZE_TEST[::-1]), interpolation=cv2.INTER_LINEAR
            )
            # patch, scale = preprocess(patch, self.cfg.INPUT.SIZE_TEST[::-1])

            # plt.figure()
            # plt.imshow(patch)
            # plt.show()

            # Make shape with a new batch dimension which is adapted for network input
            patch = torch.as_tensor(patch.astype("float32").transpose(2, 0, 1))
            patch = patch.to(device=self.device).half()

            patches.append(patch)

            if (d + 1) % self.batch_size == 0:
                patches = torch.stack(patches, dim=0)
                batch_patches.append(patches)
                patches = []

        if len(patches):
            patches = torch.stack(patches, dim=0)
            batch_patches.append(patches)

        features = np.zeros((0, 2048))
        # features = np.zeros((0, 768))

        for patches in batch_patches:

            # Run model
            patches_ = torch.clone(patches)
            pred = self.model(patches)
            pred[torch.isinf(pred)] = 1.0

            feat = postprocess(pred)

            nans = np.isnan(np.sum(feat, axis=1))
            if np.isnan(feat).any():
                for n in range(np.size(nans)):
                    if nans[n]:
                        # patch_np = patches[n, ...].squeeze().transpose(1, 2, 0).cpu().numpy()
                        patch_np = patches_[n, ...]
                        patch_np_ = torch.unsqueeze(patch_np, 0)
                        pred_ = self.model(patch_np_)

                        patch_np = torch.squeeze(patch_np).cpu()
                        patch_np = torch.permute(patch_np, (1, 2, 0)).int()
                        patch_np = patch_np.numpy()

                        plt.figure()
                        plt.imshow(patch_np)
                        plt.show()

            features = np.vstack((features, feat))

        return features



import supervisely as sly
from logging import Logger


class Tracker3DInterface:
    def __init__(
        self,
        context,
        api,
    ):
        self.api: sly.Api = api
        self.logger: Logger = api.logger
        self.frame_index = context["frameIndex"]
        self.frames_count = context["frames"]

        self.track_id = context["trackId"]
        self.dataset_id = context["datasetId"]
        self.pc_ids = (
            context["pointCloudIds"]
            if context["direction"] == "forward"
            else context["pointCloudIds"][::-1]
        )
        self.figure_ids = context["figureIds"]
        self.object_ids = context["objectIds"]
        self.direction = context["direction"]

        self.stop = (
            len(self.pc_ids)
            + (len(self.figure_ids) * len(self.pc_ids))
            + (len(self.figure_ids) * self.frames_count)
        )
        self.global_pos = 0
        self.global_stop_indicatior = False
        self.pc_dir = "./pointclouds/"

        self.geometries = []
        self.frames_indexes = []

        self.add_frames_indexes()
        self.add_geometries()

        self.logger.info("Tracker 3D interface initialized")

    def add_geometries(self):
        for figure_id in self.figure_ids:
            figure = self.api.pointcloud.figure.get_info_by_id(figure_id)
            geometry = sly.deserialize_geometry(figure.geometry_type, figure.geometry)
            self.geometries.append(geometry)

    def add_frames_indexes(self):
        total_frames = len(self.api.pointcloud_episode.get_frame_name_map(self.dataset_id))
        cur_index = self.frame_index

        while 0 <= cur_index < total_frames and len(self.frames_indexes) < self.frames_count:
            self.frames_indexes.append(cur_index)
            cur_index += 1 if self.direction == "forward" else -1

    def _notify(
        self,
        stop: bool = False,
        task: str = "not defined",
    ):
        self.global_pos += 1

        if stop:
            pos = self.stop
        else:
            pos = self.global_pos

        self.logger.debug(f"Task: {task}")
        self.logger.debug(f"Notification status: {pos}/{self.stop}")

        self.global_stop_indicatior = self.api.pointcloud_episode.notify_progress(
            self.track_id,
            self.dataset_id,
            self.pc_ids,
            pos,
            self.stop,
        )

        if self.global_stop_indicatior and self.global_pos < self.stop:
            self.logger.info("Task stoped by user")

    def add_cuboid_on_frame(self, pcd_id, object_id, cuboid_json, track_id):
        self.api.pointcloud_episode.figure.create(
            pcd_id, object_id, cuboid_json, "cuboid_3d", track_id
        )
        self._notify(task="add geometry on frame")



from supervisely.app.widgets import Container, OneOf, Text, Flexbox
import supervisely.nn.inference.instance_segmentation.dashboard.preview as preview


oneof_block = OneOf(preview.image_source)

preview_image_layout = Container(
    [preview.image_source_field, oneof_block],
)



# coding: utf-8

from collections import namedtuple

from supervisely.sly_logger import logger
from supervisely.metric.common import log_head, log_line, TOTAL_GROUND_TRUTH
from supervisely.metric.matching import filter_labels_by_name, match_labels_by_iou
from supervisely.metric.metric_base import MetricsBase

import numpy as np

MATCHES = 'matches'
AP = 'average-precision'

MatchWithConfidence = namedtuple('MatchWithConfidence', ['is_correct', 'confidence'])


class MAPMetric(MetricsBase):

    def __init__(self, class_mapping, iou_threshold, confidence_tag_name='confidence', confidence_threshold=0.0):
        if len(class_mapping) < 1:
            raise RuntimeError('At least one classes pair should be defined!')
        self._gt_to_pred_class_mapping = class_mapping.copy()
        self._pred_to_gt_class_mapping = {
            pred_name: gt_name for gt_name, pred_name in self._gt_to_pred_class_mapping.items()}
        if len(self._pred_to_gt_class_mapping) != len(self._gt_to_pred_class_mapping):
            raise ValueError('Mapping from ground truth to prediction class name is not 1-to-1: multiple ground truth '
                             'classes are mapped to the same prediction class. This is not supported for mAP '
                             'calculations.')

        self._iou_threshold = iou_threshold
        self._confidence_tag_name = confidence_tag_name
        self._confidence_threshold = confidence_threshold
        self._counters = {
            gt_cls: {MATCHES: [], TOTAL_GROUND_TRUTH: 0} for gt_cls in self._gt_to_pred_class_mapping.keys()}

    def _get_confidence_value(self, label):
        confidence_tag = label.tags.get(self._confidence_tag_name, None)
        return confidence_tag.value if confidence_tag is not None else None

    def add_pair(self, ann_gt, ann_pred):
        labels_gt = filter_labels_by_name(ann_gt.labels, self._gt_to_pred_class_mapping)
        all_labels_pred = [label for label in filter_labels_by_name(ann_pred.labels, self._pred_to_gt_class_mapping)]
        labels_pred = []
        for label in all_labels_pred:
            label_confidence = self._get_confidence_value(label)
            if label_confidence is None:
                logger.warn(f'Found a label with class {label.obj_class.name!r} that does not have a '
                            f'{self._confidence_tag_name!r} tag attached. Skipping this object for metric computation.')
            elif label_confidence >= self._confidence_threshold:
                labels_pred.append(label)
        match_result = match_labels_by_iou(labels_1=labels_gt, labels_2=labels_pred, img_size=ann_gt.img_size,
                                           iou_threshold=self._iou_threshold)
        for match in match_result.matches:
            gt_class = match.label_1.obj_class.name
            label_pred = match.label_2
            self._counters[gt_class][MATCHES].append(
                MatchWithConfidence(is_correct=(label_pred.obj_class.name == self._gt_to_pred_class_mapping[gt_class]),
                                    confidence=self._get_confidence_value(label_pred)))
        # Add unmatched predictions to the list as false positive matches.
        for umatched_pred in match_result.unmatched_labels_2:
            gt_class = self._pred_to_gt_class_mapping[umatched_pred.obj_class.name]
            self._counters[gt_class][MATCHES].append(
                MatchWithConfidence(is_correct=False, confidence=self._get_confidence_value(umatched_pred)))

        for label_1 in labels_gt:
            self._counters[label_1.obj_class.name][TOTAL_GROUND_TRUTH] += 1

    @staticmethod
    def _calculate_average_precision(gt_class, pred_class, pair_counters):
        if len(pair_counters[MATCHES]) == 0 or all(match.is_correct == False for match in pair_counters[MATCHES]):
            logger.warning('No matching samples for pair {!r} <-> {!r} have been detected. '
                           'MAP value for this pair will be set to 0.'.format(gt_class, pred_class))
            return 0

        sorted_matches = sorted(pair_counters[MATCHES], key=lambda match: match.confidence, reverse=True)
        correct_indicators = [int(match.is_correct) for match in sorted_matches]
        total_correct = np.cumsum(correct_indicators)
        recalls = total_correct / pair_counters[TOTAL_GROUND_TRUTH]
        precisions = total_correct / (np.arange(len(correct_indicators)) + 1)
        anchor_precisions = []
        for anchor_recall in np.linspace(0, 1, 11):
            points_above_recall = (recalls >= anchor_recall)
            anchor_precisions.append(np.max(precisions[points_above_recall]) if np.any(points_above_recall) else 0)
        return np.mean(anchor_precisions)

    def get_metrics(self):  # Macro-evaluation
        result = {gt_class:
                      {AP: self._calculate_average_precision(gt_class, self._gt_to_pred_class_mapping[gt_class],
                                                             pair_counters)}
                  for gt_class, pair_counters in self._counters.items()}
        return result

    @staticmethod
    def average_per_class_avg_precision(per_class_metrics):
        return np.mean([class_metrics[AP] for class_metrics in per_class_metrics.values()])

    def get_total_metrics(self):
        return {AP: self.average_per_class_avg_precision(self.get_metrics())}

    def log_total_metrics(self):
        log_line()
        log_head(' Result metrics values for {} IoU threshold '.format(self._iou_threshold))

        classes_values = self.get_metrics()
        for i, (cls_gt, pair_values) in enumerate(classes_values.items()):
            average_precision = pair_values[AP]
            log_line()
            log_head(' Results for pair of classes <<{} <-> {}>>  '.format(cls_gt,
                                                                           self._gt_to_pred_class_mapping[cls_gt]))
            logger.info('Average Precision (AP): {}'.format(average_precision))

        log_line()
        log_head(' Mean metrics values ')
        logger.info('Mean Average Precision (mAP): {}'.format(self.get_total_metrics()[AP]))
        log_line()
